%%
%  IBM 4758  CCA API
%
%  Recomendation 2, Role C:
%
%  The intuder has acces to Encipher, Decipher, KeyImport, KeyExport,
%  KeyTranslate (which is subsumed by KeyImport and KeyExport), and
%  KPI-Last. However, the intruder has NO access to the
%  pin-derivation-key to be importer ({pdkey}_{kek+pin}).
%%

    fun enc/2.      % symmetric encryption
    fun penc/2.     % public key encryption

    % types of keys:
    % fun dat/0.    % data keys  (data = 0)
    fun imp/0.      % import keys
    fun exp/0.      % export keys
    fun pin/0.      % key derivation keys
    fun kp/0.       % partial key type
    fun pan/0.      % an account number.

    fun kp12/0.     % a partial key
    fun kp12a/0.    % a partial key
    fun kp3/0.      % a partial key - the part of the intruder

    fun km/0.       % the security module's master key 

    fun c/0.        % some value known by the intruder

    fun ukey/1.      % key(Type) -- a key of type Type generated by the module or guessed.
    fun ukek1/0.   
    fun ukek2/0.   

    fun pdkey/0.    % a Pin Derivation Key

    fun sec/0.

    proverif 'param selFun = Nounifset'.

% dominating [imp, exp, pin, kp, kp12 ].

%%%%%%%%%%%% Auxiliary Prolog predicates

    type(0).   % data=0
    type(imp).
    type(exp).
    type(pin).
    
%%%%%%%%%%%% THE INITIAL KNOWLEDGE %%%%%%%%

    rule c.
    % rule dat.  % dat=0
    rule imp.
    rule exp.
    rule pin.
    rule kp.

    rule pan.
    rule kp3.
    % rule enc(pdkey, kp12+kp3+pin).  % the key to be imported
    rule enc(pdkey, km+pin).

    %%% Key Generating/Guessing

    rule enc(ukey(0),   km+0).
    rule enc(ukey(imp), km+imp).
    rule enc(ukey(exp), km+exp).
    rule enc(ukey(pin), km+pin).

%%%%%%%%%%%% INTRUDER RULES %%%%%%%%%%%%%%%

    rule  X, K -> enc(X,K).
    rule  enc(X,K), K -> X.

%%%%%%%%%%%% PROTOCOL RULES %%%%%%%%%%%%%%%

%%% Encipher
    rule  X, enc(Key, km+0)  ->  enc(X, Key).

%%% Decipher
    rule  enc(X, Key), enc(Key, km+0)  ->  X.

%%% Key Import
    rule  enc(Key, KEK+Type), Type, enc(KEK, km+imp)  ->  enc(Key, km+Type)
    :- type(Type).

%%% Key Export
    rule  enc(Key, km+Type), Type, enc(KEK, km+exp)  ->  enc(Key, KEK+Type)
    :- type(Type).


%%% Key Part Import

    % KPI-First + KPI-Add/Middle (run by some honest agents):
    rule  enc( kp12, km+kp+imp ).
    % rule  enc( kp12, km+kp+exp ).
    % rule  enc( kp12, km+kp+pin ).
    % rule  enc( kp12a, km+kp+0 ).
    
    % KPI-Last:   X, Type, enc(Y, km+kp+Type)  ->  enc(X+Y, km+Type).
    % unfolded:
    
    rule  X, Type, km+kp+Type  ->  enc(X, km+Type)
    :- type(Type).

    rule  X, imp  ->  enc(X+kp12, km+imp).
    % rule  X, exp  ->  enc(X+kp12, km+exp).
    % rule  X, pin  ->  enc(X+kp12, km+pin).
    % rule  X, 0    ->  enc(X+kp12a, km+0).

    % rule  X, Type ->  enc(X+ukek1, km+Type)  :- type(Type).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rule  enc(pdkey, km)  -> sec.
rule  pdkey -> sec.
rule  enc(pan, pdkey) -> sec.
rule  kp12+kp3 -> sec.
rule  ukek1 -> sec.
rule  ukek2 -> sec.
rule  ukey(_) -> sec.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
query sec.

