% Functors:
    fun enc/2. 
    fun h/2.   
    fun key/1. 
    fun k/2.
    fun r/1.   

    fun a/0.   
    fun b/0.    
    fun end/0.
    fun n/1.
    fun secret/0.

% Auxiliaray predicates
    principal(a).
    principal(b).

%%% The attacker (encryption, decryption, and keyed hashing)

    rule  X, Y -> enc(X,Y).
    rule  Y, enc(X,Y) -> X.
    rule  key(X), Y -> h(key(X),Y).

%%% The initial knowledge:

    rule a.
    rule b.
    rule n(b).
    rule key(b).

%%% Protocol rules:
%
% In the descriptions h_K(M) stands for (h(k,M),M).
%

%%% A. Rules to build a request:
  
% init rule:  h_KA(A, B, Na, end)

    rule    (h(key(A),(A,B,n(A),end)), (A,B,n(A),end))
    :- 
        principal(A), principal(B).

% continuation rule (a rule to making a chain of requests):
%     h_KA(A,B,Na,Rest)  -->   h_KB(B, C, Nb, h_KA(...) ) 

    rule    (H,(A,B,NA,Rest)) 
            -> ( h(key(B), (B,C,n(B), (H,(A,B,NA,Rest)))), (B,C,n(B), (H,(A,B,NA,Rest))) )
    :- 
        principal(A), principal(B), principal(C).
    

%%% B. Rules modelthe request processing:

% initialisation of recursion  

    rule  (H, (C, a, NC, Rest))  ->  r( (H, (C, a, NC, Rest)) ). 

% recurvisve rule:  
%     r( h_Ki(Pi, Pj, x, y) )  -->  r(y)  

    rule  r(( h(key(Pi), (Pi, Pj, X, Y)), (Pi, Pj, X, Y) ))  ->  r(Y).

% send rule (1):
%    r( h_Ki(Pi, Pj, x, h_Kl(Pl, Pi, x', y)) -->  i((Kij,Pj,x)*Ki), i((Kil,Pl,x)*Ki)  

    rule  r( ( h(key(Pi), (Pi,Pj,X,(
                                  h(key(Pl), (Pl,Pi,X1,Y)),
                                             (Pl,Pi,X1,Y)
                              ))), 
                       (Pi,Pj,X,(
                                  h(key(Pl), (Pl,Pi,X1,Y)),
                                             (Pl,Pi,X1,Y)
                              )) ) )
 
         ->  ( 
                 (Pi, Pj,  k(Pi,Pj) + h(key(Pi),X), enc((Pi,Pj,X), k(Pi,Pj))),
                 (Pi, Pl,  k(Pl,Pi) + h(key(Pi),X), enc((Pi,Pl,X), k(Pl,Pi)))
             ) 
    :- 
        principal(Pi), principal(Pj), principal(Pl).

% send rule (2):   
%    r( h_Ki(Pi, Pj, X, end))  -->  i((Kij,Pj,X)*Ki)  

    rule       
        r(( h(key(Pi), (Pi,Pj,X,end)), (Pi,Pj,X,end) ))  ->
            (Pi, Pj,  (k(Pi,Pj)+h(key(Pi),X)), enc((Pi,Pj,X), k(Pi,Pj)))
    :- 
        principal(Pi), principal(Pj).


% The secret:
query k(a,a).
