%
% Symmetric key distribution using Smart Cards, by Shoup and Rubin.
%

%%% functions and constants:
    fun enc/2.  % symmetric key cryptograpy 
    fun key/1.  % key(a) -- the key of a 
    fun kc/1.   % k(a,c) -- the key shared by a and c

    fun  a/0.
    fun  b/0.
    fun  s/0.
    fun  l0/0.
    fun  l1/0.
    fun  l2/0.
    fun  na/2.
    fun  n/2.

    fun  secret/0.
    fun  sec2/0.
    fun  sec3/0.

%%% Auxiliary predicates:
    princ(a).
    princ(b).
    p_Kab(A, B, KAB)  :-  KAB = enc((A,l0),key(B)). 
    p_Pab(A, B, PAB)  :-  p_Kab(A,B,KAB), PAB = KAB+enc((B,l1), key(A)). 

%%% Initial knowledge
    rule a.
    rule b.
    rule s.
    rule l0.
    rule l1.
    rule l2.
    rule key(b).
    rule kc(b).

%%% The intruder rules :

    rule X, Y -> enc(X,Y).
    rule enc(X,Y), Y -> X.

%%% Protocol rules:

    %% 1 -- we skip it: the intruder learns nothing

    %% 2 (the responce of the server)

    rule  (PAB, enc((PAB, B, l2), key(A)))
     :- princ(A), princ(B), p_Pab(A,B,PAB).

    %% 3 -- skip
    
    %% 4 -- only for b (the connection between a and Ca is secure)

    rule  na(b,a), enc( (na(b,a), l1, l1), kc(b)).
    rule  na(b,b), enc( (na(b,b), l1, l1), kc(b)).

    %% 5  
        (
          rule  (A,na(A,B)) 
        )
          :- princ(A), princ(B).

    %% 6 -- skip
    
    %% 7 -- only for b (the connection between a and Ca is secure)

    rule
        A, NA  ->
            n(b,A), 
            enc( (n(b,A), l0, l0), KAB),
            enc( (NA, n(b,A), l1), KAB),
            enc( (n(b,A), l0, l1), KAB)
    :- princ(A), p_Kab(A,b,KAB).


    %% 8
    
    rule (A,NA) -> (n(B,A), enc((NA, n(B,A), l1), KAB))
    :- 
        princ(A), princ(B), p_Kab(A,B,KAB).


    %% 9 -- skip (the connection between a and Ca is secure
    
    %% 10 -- only for b (A=b)

    rule
        B, NA, NB, PAB, 
        enc((PAB, B, l2), key(A)),
        enc((NA, NB, l1), KAB),
        enc((NB, l0, l1), KAB)
          ->  enc((NB, l0, l0), KAB),
              enc((NB, l0, l1), KAB)
    :- 
        A=b, princ(B), p_Pab(A,B,PAB), p_Kab(A,B,KAB).

    %% 11 

    rule  (NB, enc((na(A,B), NB, l1), KAB) ) 
          ->  (enc((NB,l0,l1),  KAB))
    :- 
        princ(A), princ(B), p_Kab(A,B,KAB).
      

%%% The secret:

query  enc( (n(a,a), l0, l0),  enc((a,l0),key(a)) ).

